﻿//// Parts of this script are derived from "Simple procedural terrain in Unity" written in C# by Scrawk.
//// Javascript version with additions and modifications by Virtue One.

//// NOTES: 
//// 1) 'Perlin.cs' needs to be placed into the "Plugins" folder, while this script needs to be in the root folder, or loaded first (Provided by Scrawk).

var waterObj:GameObject;
var cloudsObj:GameObject;

// Terrain Prototypes, materials, textures, and other vars...
var heightmapCustom:Texture2D;
private var sealevel:float;
private var snowcapheight:float;
var TerrainMat:Material;
var shiny:boolean = false;
var m_splat0:Texture2D;
var m_splat1:Texture2D;
var m_splat2:Texture2D;
var m_splat3:Texture2D;
var m_normMap0:Texture2D;
var m_normMap1:Texture2D;
var m_normMap2:Texture2D;
var m_normMap3:Texture2D;
var m_splatTileSize0:float = 32.0f;
var m_splatTileSize1:float = 32.0f;
var m_splatTileSize2:float = 32.0f;
var m_splatTileSize3:float = 32.0f;
var m_detail0:Texture2D;
var m_detail1:Texture2D;
var m_detail2:Texture2D;
var m_trees:GameObject[];
//Noise settings. A higher frq will create larger scale details. Each seed value will create a unique look
private var m_groundSeed:int = 0;
var m_groundFrq:float = 800.0f;
private var m_mountainSeed:int = 1;
var m_mountainFrq:float = 1200.0f;
private var m_treeSeed:int = 2;
var m_treeFrq:float = 400.0f;
private var m_detailSeed:int = 3;
var m_detailFrq:float = 100.0f;
//Terrain settings
private var m_pixelMapError:float = 6.0f; //A lower pixel error will draw terrain at a higher Level of detail but will be slower
private var m_baseMapDist:float = 1000.0f; //The distance at which the low res base map will be drawn. Decrease to increase performance
//Terrain data settings
private var m_heightMapSize:int; //Higher number will create more detailed height maps
private var m_alphaMapSize:int; //This is the control map that controls how the splat textures will be blended
var m_terrainSize:int = 2048;
var m_terrainHeight:int = 512;
var m_detailMapSize:int = 512; //Resolutions of detail (Grass) layers
//Tree settings
var m_treeSpacing:int = 32; //spacing between trees
var m_treeVolume:int = 1500;
private var m_treeDistance:float = 2000.0f; //The distance at which trees will no longer be drawn
private var m_treeBillboardDistance:float = 2000.0f; //The distance at which trees meshes will turn into tree billboards
private var m_treeCrossFadeLength:float = 20.0f; //As trees turn to billboards there transform is rotated to match the meshes, a higher number will make this transition smoother
private var m_treeMaximumFullLODCount:int = 400; //The maximum number of trees that will be drawn in a certain area. 
//Detail settings
private var detailMode:DetailRenderMode;
var m_detailObjectDistance:int = 500; //The distance at which details will no longer be drawn
var m_detailObjectDensity:float = 4.0f; //Creates more dense details within patch
var m_detailResolutionPerPatch:int = 32; //The size of detail patch. A higher number may reduce draw calls as details will be batch in larger patches
var m_wavingGrassStrength:float = 0.4f;
var m_wavingGrassAmount:float = 0.2f;
var m_wavingGrassSpeed:float = 0.4f;
var m_wavingGrassTint:Color = Color.white;
var m_grassHealthyColor:Color = Color.white;
var m_grassDryColor:Color = Color.white;
var grassleveloffset:float = 2.5f;
var m_grassHeightMax:float = 5.0f;

var m_groundNoise:PerlinNoise;
var m_mountainNoise:PerlinNoise;
var m_treeNoise:PerlinNoise;
var m_detailNoise:PerlinNoise;
private var m_terrain:Terrain;
var m_splatPrototypes:SplatPrototype[];
var m_treeProtoTypes:TreePrototype[];
var m_detailProtoTypes:DetailPrototype[];

function Start() {
	m_groundSeed = Random.Range(1000,3000); //2208
	m_mountainSeed = Random.Range(1,800); //5
	m_detailSeed = m_mountainSeed;
	m_treeSeed = Random.Range(1,8);
	if (m_mountainFrq==0) m_mountainFrq = Random.Range(80,350);  //8-300
	m_groundNoise = new PerlinNoise(m_groundSeed);
	m_mountainNoise = new PerlinNoise(m_mountainSeed);
	m_treeNoise = new PerlinNoise(m_treeSeed);
	m_detailNoise = new PerlinNoise(m_detailSeed);
	m_baseMapDist = 2000.0f;
	if (!m_detailObjectDistance) m_detailObjectDistance = 500;
	snowcapheight = m_terrainHeight/3.5f;
	
	if (heightmapCustom) {
		m_heightMapSize = heightmapCustom.width;
	} else {
		if (!m_heightMapSize) {
			m_heightMapSize = m_terrainSize/2+1;
		}
	}
	m_alphaMapSize = m_heightMapSize-1;
	
	
	//// set position and scale for water object.
	if (waterObj) {
		sealevel = waterObj.transform.position.y+6;
		waterObj.transform.position = Vector3(m_terrainSize/2,waterObj.transform.position.y,m_terrainSize/2);
		waterObj.transform.localScale = Vector3(m_terrainSize/50,1,m_terrainSize/50);
	} else {
		sealevel = 0;
	}
	
	//// set position and scale for clouds object.
	if (cloudsObj) {
		var cloudsSize:float = m_terrainSize*1.5;
		cloudsObj.transform.position = Vector3(m_terrainSize/2,cloudsObj.transform.position.y,m_terrainSize/2);
		cloudsObj.transform.localScale = Vector3(cloudsSize,cloudsObj.transform.localScale.y,cloudsSize);
	}
	
	if(!Mathf.IsPowerOfTwo(m_heightMapSize-1))
	{
		Debug.Log("TerrianGenerator::Start - height map size must be pow2+1 number");
		m_heightMapSize = Mathf.ClosestPowerOfTwo(m_heightMapSize)+1;
	}
	
	if(!Mathf.IsPowerOfTwo(m_alphaMapSize))
	{
		Debug.Log("TerrianGenerator::Start - Alpha map size must be pow2 number");
		m_alphaMapSize = Mathf.ClosestPowerOfTwo(m_alphaMapSize);
	}
	
	if(!Mathf.IsPowerOfTwo(m_detailMapSize))
	{
		Debug.Log("TerrianGenerator::Start - Detail map size must be pow2 number");
		m_detailMapSize = Mathf.ClosestPowerOfTwo(m_detailMapSize);
	}
	
	if(m_detailResolutionPerPatch < 8)
	{
		Debug.Log("TerrianGenerator::Start - Detail resolution per patch must be >= 8, changing to 8");
		m_detailResolutionPerPatch = 8;
	}
	
	var htmap = new float[m_heightMapSize,m_heightMapSize];
	//m_terrain = new Terrain;
	CreateProtoTypes();
	FillHeights(htmap);
	
	var terrainData:TerrainData = new TerrainData();
	terrainData.heightmapResolution = m_heightMapSize;
	terrainData.SetHeights(0, 0, htmap);
	terrainData.size = new Vector3(m_terrainSize, m_terrainHeight, m_terrainSize);
	terrainData.splatPrototypes = m_splatPrototypes;
	terrainData.treePrototypes = m_treeProtoTypes;
	terrainData.detailPrototypes = m_detailProtoTypes;
	terrainData.thickness = 5.0f;
	
	FillAlphaMap(terrainData);
	m_terrain = Terrain.CreateTerrainGameObject(terrainData).GetComponent(Terrain);
	m_terrain.transform.position = new Vector3(0, 0, 0); // CUSTOM TERRAIN POSITION
	m_terrain.heightmapPixelError = m_pixelMapError;
	m_terrain.basemapDistance = m_baseMapDist;
	
	m_terrain.castShadows = true; //disable this for better frame rate
	m_terrain.materialType = Terrain.MaterialType.Custom;
	m_terrain.materialTemplate = TerrainMat;
	if (shiny) {
		m_terrain.legacyShininess = 0.5f;
		m_terrain.legacySpecular = new Color(0.5f,0.5f,0.5f);
	} else {
		//m_terrain[0].materialType = Terrain.MaterialType.BuiltInLegacyDiffuse;
		//m_terrain[0].materialType = Terrain.MaterialType.BuiltInLegacySpecular;
		m_terrain.legacyShininess = 0.0f;
		m_terrain.legacySpecular = new Color(0.5f,0.5f,0.5f);
	}
	FillTreeInstances(m_terrain);
	FillDetailMap(m_terrain);
	
}
	
function CreateProtoTypes()	{
		m_splatPrototypes = new SplatPrototype[4];
		
		m_splatPrototypes[0] = new SplatPrototype();
		m_splatPrototypes[0].texture = m_splat0;
		m_splatPrototypes[0].tileSize = new Vector2(m_splatTileSize0, m_splatTileSize0);
		m_splatPrototypes[0].normalMap = m_normMap0;
		
		m_splatPrototypes[1] = new SplatPrototype();
		m_splatPrototypes[1].texture = m_splat1;
		m_splatPrototypes[1].tileSize = new Vector2(m_splatTileSize1, m_splatTileSize1);
		m_splatPrototypes[1].normalMap = m_normMap1;
		
		m_splatPrototypes[2] = new SplatPrototype();
		m_splatPrototypes[2].texture = m_splat2;
		m_splatPrototypes[2].tileSize = new Vector2(m_splatTileSize2, m_splatTileSize2);
		m_splatPrototypes[2].normalMap = m_normMap2;
		
		m_splatPrototypes[3] = new SplatPrototype();
		m_splatPrototypes[3].texture = m_splat3;
		m_splatPrototypes[3].tileSize = new Vector2(m_splatTileSize3, m_splatTileSize3);
		m_splatPrototypes[3].normalMap = m_normMap3;
		
		m_treeProtoTypes = new TreePrototype[m_trees.length];
		for (var treeindex:int = 0; treeindex<m_trees.length;treeindex++) {
			m_treeProtoTypes[treeindex] = new TreePrototype();
			m_treeProtoTypes[treeindex].prefab = m_trees[treeindex];
			m_treeProtoTypes[treeindex].bendFactor = 1.2f;
		}
		
		m_detailProtoTypes = new DetailPrototype[3];
		// grass values...
		if (m_grassHeightMax<1.0f) m_grassHeightMax = 1.0f;
		var maxW:float;
		maxW = m_grassHeightMax+0.5f;
		m_detailProtoTypes[0] = new DetailPrototype();
		m_detailProtoTypes[0].prototypeTexture = m_detail0;
		m_detailProtoTypes[0].renderMode = detailMode;
		m_detailProtoTypes[0].healthyColor = m_grassHealthyColor;
		m_detailProtoTypes[0].dryColor = m_grassDryColor;
		m_detailProtoTypes[0].minHeight = 1.5f;
		m_detailProtoTypes[0].minWidth = 1.5f;
		m_detailProtoTypes[0].maxHeight = maxW;
		m_detailProtoTypes[0].maxWidth = maxW;
		
		maxW = m_grassHeightMax;
		m_detailProtoTypes[1] = new DetailPrototype();
		m_detailProtoTypes[1].prototypeTexture = m_detail1;
		m_detailProtoTypes[1].renderMode = detailMode;
		m_detailProtoTypes[1].healthyColor = m_grassHealthyColor;
		m_detailProtoTypes[1].dryColor = m_grassDryColor;
		m_detailProtoTypes[1].minHeight = 1.2f;
		m_detailProtoTypes[1].minWidth = 1.2f;
		m_detailProtoTypes[1].maxHeight = maxW;
		m_detailProtoTypes[1].maxWidth = maxW;
		
		maxW = m_grassHeightMax+1.0f;
		m_detailProtoTypes[2] = new DetailPrototype();
		m_detailProtoTypes[2].prototypeTexture = m_detail2;
		m_detailProtoTypes[2].renderMode = detailMode;
		m_detailProtoTypes[2].healthyColor = m_grassHealthyColor;
		m_detailProtoTypes[2].dryColor = m_grassDryColor;
		m_detailProtoTypes[2].minHeight = 1.8f;
		m_detailProtoTypes[2].minWidth = 1.8f;
		m_detailProtoTypes[2].maxHeight = maxW;
		m_detailProtoTypes[2].maxWidth = maxW;
}
	
function FillHeights(htmap:float[,]) {
	var pix:Color32[] = new Color32[m_heightMapSize*m_heightMapSize];
	var hprev:float = 0.0f;
	if (heightmapCustom) pix = heightmapCustom.GetPixels32();
	var ratio:float = (m_terrainSize*1.0f)/(m_heightMapSize*1.0f);
	//var ratio:float = m_terrainSize/m_heightMapSize;
	var v:float = 1.0f / 15.0f;
	var kernel = [[ v, v, v ],[ v, v, v ],[ v, v, v ]];
	for(var x:int = 0; x < m_heightMapSize; x++)
	{
		for(var z:int = 0; z < m_heightMapSize; z++)
		{
			var terrval:float = 0.0f;
			if (heightmapCustom) {
				if (m_heightMapSize>513) {
					var sum:float = 0; // Kernel sum for this pixel - this process used for a smoothing effect.
				      for (var ky:int = -1; ky <= 1; ky++) {
				        for (var kx:int = -1; kx <= 1; kx++) {
				          //// Calculate the adjacent pixel for this kernel point
				          if ((z+ky)*m_heightMapSize+(x+kx) > -1 && (z+ky)*m_heightMapSize+(x+kx) < pix.Length) { 
				          	var val:float = pix[(z+ky)*m_heightMapSize+(x+kx)].r;
				          	sum += kernel[ky+1][kx+1] * val; // Multiply adjacent pixels based on the kernel values
				          }
				        }
				      }
				    var norm:float = sum / 255.0f;
					terrval = norm;
				} else {
					terrval = heightmapCustom.GetPixel(x, z).grayscale;
					terrval = (hprev+terrval)/2;
					hprev = terrval;
				}
			} else {
				var finalh:float = 0.0f;
				var r:float = 0.0f;
				var worldPosX:float = (x+1*(m_heightMapSize-1))*ratio;
				var worldPosZ:float = (z+1*(m_heightMapSize-1))*ratio;
				var mountains:float = Mathf.Max(0.0f, m_mountainNoise.FractalNoise2D(worldPosX, worldPosZ, 6, m_mountainFrq, 0.8f));
				var plain:float = m_groundNoise.FractalNoise2D(worldPosX, worldPosZ, 4, m_groundFrq, 0.1f) + 0.1f;
				//// GRADIENT HEIGHT FALLOFF....
		        //float r = plain+mountains;
			    var maxdist:float = m_heightMapSize/2.0f; //dist(0,0,width,height);
			    //// Formula for gradient (from center point)
			    var d:float = Mathf.Sqrt((x-m_heightMapSize/2.0f)*(x-m_heightMapSize/2.0f) + (z-m_heightMapSize/2.0f)*(z-m_heightMapSize/2.0f));
			    var adjustbrightness:float  = 100*(maxdist-d)/maxdist*1.5f;
			    r += adjustbrightness*0.01f;
			    ////control/clamp values
			    if (r<=0.0f) r = 0.0f;
			    if (r>=1.0f) r = 1.0f;
				finalh = plain+mountains;
				if (finalh > r) {
					finalh = r - ((finalh)-r);
					//finalh = r - ((plain+mountains)/10.0f); //testing
				}
				if (finalh<0.05f) finalh = 0.05f;
				terrval = finalh;
			}
			
			htmap[z,x] = terrval;
			//htmap[z,x] = plain+mountains;
		}
	}
}
	
function FillAlphaMap(terrainData:TerrainData) {
	var map:float[,,]  = new float[m_alphaMapSize, m_alphaMapSize, 4];
	Random.seed = 0;
    for(var x:int = 0; x < m_alphaMapSize; x++) 
	{
    	for (var z:int = 0; z < m_alphaMapSize; z++) 
		{
        // Get the normalized terrain coordinate that
        // corresponds to the the point.
        var normX:float = x * 1.0f / (m_alphaMapSize - 1);
        var normZ:float = z * 1.0f / (m_alphaMapSize - 1);
        
        // Get the steepness value at the normalized coordinate.
        var angle:float = terrainData.GetSteepness(normX, normZ);
        var height:float = terrainData.GetHeight(x,z);

		//// Steepness is given as an angle, 0..90 degrees. Divide
        //// by 90 to get an alpha blending value in the range 0..1.
        var frac:float; // was 90.0f
        
        var angle_div_amt:float = m_terrainSize/250.0f;
        if (m_terrainHeight<=100) {
        	frac = angle / 38.0f;
        } else {
        	frac = angle / (40.0f + angle_div_amt);
        }
		var heightval:float = height;
		if (height<sealevel) {
			//// fade height texture based on 0.0-1.0 scale...
			var xMax = 1.0;
			var xMin = 0.0;
			var sealevelMax = sealevel-8;
			var sealevelMin = sealevel;
			var percent = (height - sealevelMin) / (sealevelMax - sealevelMin);
			heightval = percent * (xMax - xMin) + xMin;
			} else {
               	heightval = 0.0f;
			}
			var snowheightval:float  = 0.0f;
			var snowcapheightRand:float = Random.Range(snowcapheight-10.0f,snowcapheight+10.0f);
            if (height>snowcapheightRand && snowcapheight!=0) {
               	////map[z, x, 3] = heightval;
				//// fade height texture based on 0.0-1.0 scale...
				var snowcapMax = snowcapheightRand+45.0;
				var snowcapMin = snowcapheightRand;
				percent = (height - snowcapMin) / (snowcapMax - snowcapMax);
				snowheightval = percent * (xMax - xMin) + xMin;
			} else {
				snowheightval = 0.0f;
			}
			map[z, x, 0] = frac - heightval - snowheightval;
			//map[z, x, 1] = 1.0f - frac - height*1.5f; //testing
			map[z, x, 1] = 1.0f - frac - heightval - snowheightval;
			map[z, x, 2] = heightval - snowheightval;
			map[z, x, 3] = snowheightval;
        }
	}
	terrainData.alphamapResolution = m_alphaMapSize;
	terrainData.SetAlphamaps(0, 0, map);
}

function FillTreeInstances(terrain:Terrain) {
	Random.seed = 0;
	for(var x:int = 0; x < m_terrainSize; x++) {
		for (var z:int = 0; z < m_terrainSize; z++) {
			//// using method below to control tree volume, instead of 'spacing' originally used.
			var randomplacement:int = Random.Range(1,m_treeVolume);
			if (randomplacement==1) {
				var unit:float = 1.0f / (m_terrainSize - 1);
				
				var offsetX:float = Random.value * unit * m_treeSpacing;
				var offsetZ:float = Random.value * unit * m_treeSpacing;
				
				var normX:float = x * unit + offsetX;
				var normZ:float = z * unit + offsetZ;
				
				// Get the steepness value at the normalized coordinate.
				var angle:float = terrain.terrainData.GetSteepness(normX, normZ);
	                
				// Steepness is given as an angle, 0..90 degrees. Divide
				// by 90 to get an alpha blending value in the range 0..1.
				var frac:float = angle / 45.0f; //// 45 used for sharper result.
				
				if(frac < 0.5f) { //make sure tree are not on steep slopes 
					var worldPosX:float = x+1*(m_terrainSize-1);
					var worldPosZ:float = z+1*(m_terrainSize-1);
					var noise:float = m_treeNoise.FractalNoise2D(worldPosX, worldPosZ, 3, m_treeFrq, 1.0f);
					var ht:float = terrain.terrainData.GetInterpolatedHeight(normX, normZ);
					var maxht:float = m_terrainHeight/3.5f;
					if (maxht<40.0f) maxht = 40.0f;
					if(noise > 0.0f && ht < m_terrainHeight*0.4f && ht > sealevel+0.25f && ht < maxht) {
						var temp:TreeInstance = new TreeInstance();
						temp.position = new Vector3(normX,ht,normZ);
						// choose random index from tree array and assign to prototypes...
						var randindex:int = Random.Range(0, m_trees.length);
						temp.prototypeIndex = randindex;
						var rad:float = Random.Range(0.0f,360.0f) * Mathf.Deg2Rad;
						temp.rotation = rad;
						// random scale, gets larger for first 3 trees
						var scalemax:float = 1.2f;
						var scalemin:float = 1.0f;
						if (randindex==0) {
							scalemin = 1.2f;
							scalemax = 1.6f;
						}
						if (randindex==1) {
							scalemin = 1.7f;
							scalemax = 2.3f;
						}
						if (randindex==2) {
							scalemin = 2.5f;
							scalemax = 3.0f;
						}
						var randsize:float = Random.Range(scalemin,scalemax);
						temp.widthScale = randsize;
						temp.heightScale = randsize;
						temp.color = Color.white;
						temp.lightmapColor = Color.white;
						terrain.AddTreeInstance(temp); //apply
					}
				}
			} // limiter
		}
	}
	//other tree vars...
	terrain.treeDistance = m_treeDistance;
	terrain.treeBillboardDistance = m_treeBillboardDistance;
	terrain.treeCrossFadeLength = m_treeCrossFadeLength;
	terrain.treeMaximumFullLODCount = m_treeMaximumFullLODCount;
}
	
function FillDetailMap(terrain:Terrain) {
	//each layer is drawn separately so if you have a lot of layers your draw calls will increase
	//limited to 3.
	var detailMap0:int[,] = new int[m_detailMapSize,m_detailMapSize];
	var detailMap1:int[,] = new int[m_detailMapSize,m_detailMapSize];
	var detailMap2:int[,] = new int[m_detailMapSize,m_detailMapSize];
	
	var ratio:float = (m_terrainSize*1.0f)/(m_detailMapSize*1.0f); //HERE
	//var ratio:float = m_terrainSize/m_detailMapSize;
	
	Random.seed = 0;
	
	for(var x:int = 0; x < m_detailMapSize; x++) {
		for (var z:int = 0; z < m_detailMapSize; z++) {
			detailMap0[z,x] = 0;
			detailMap1[z,x] = 0;
			detailMap2[z,x] = 0;
			
			var unit:float = 1.0f / (m_detailMapSize - 1);
			var normX:float = x * unit;
			var normZ:float = z * unit;
			
			// Get the steepness value at the normalized coordinate.
			var angle:float = terrain.terrainData.GetSteepness(normX, normZ);
			////float height = terrain.terrainData.GetHeight(x,z);
			
			// Steepness is given as an angle, 0..90 degrees. Divide
			// by 90 to get an alpha blending value in the range 0..1.
			var frac:float = angle / 90.0f;
			var ht:float = terrain.terrainData.GetInterpolatedHeight(normX, normZ);
			var maxht:float;
			if (snowcapheight) {
				maxht = snowcapheight;
			} else {
				maxht = m_terrainHeight/3.0f;
			}
			if (maxht<30.0f) maxht = 30.0f;
			if(frac < 0.24f && ht > sealevel-5.0f+grassleveloffset && ht < maxht) {
				var worldPosX:float = (x+1*(m_detailMapSize-1))*ratio;
				var worldPosZ:float = (z+1*(m_detailMapSize-1))*ratio;
				var rnd:float = Random.value;
					//Randomly select what freq to use
					if(rnd < 0.25f)
						m_detailFrq = 10;
					else if(rnd < 0.4f)
						m_detailFrq = 50;
					else
						m_detailFrq = 100;
				var noise:float = m_detailNoise.FractalNoise2D(worldPosX, worldPosZ, 3, m_detailFrq, 1.0f);
				if(noise > 0.0f) {
					
					//Randomly select what layer to use
					if(rnd < 0.25f)
						detailMap0[z,x] = 1;
					else if(rnd < 0.5f)
						detailMap1[z,x] = 1;
					else
						detailMap2[z,x] = 1;
				}
			}
		}
	}
	terrain.terrainData.wavingGrassStrength = m_wavingGrassStrength;
	terrain.terrainData.wavingGrassAmount = m_wavingGrassAmount;
	terrain.terrainData.wavingGrassSpeed = m_wavingGrassSpeed;
	terrain.terrainData.wavingGrassTint = m_wavingGrassTint;
	terrain.detailObjectDensity = m_detailObjectDensity;
	terrain.detailObjectDistance = m_detailObjectDistance;
	terrain.terrainData.SetDetailResolution(m_detailMapSize, m_detailResolutionPerPatch);
	
	terrain.terrainData.SetDetailLayer(0,0,0,detailMap0);
	terrain.terrainData.SetDetailLayer(0,0,1,detailMap1);
	terrain.terrainData.SetDetailLayer(0,0,2,detailMap2);
	
	/// this is required to fix bug where colliders aren't enabled...
	terrain.GetComponent(TerrainCollider).enabled = false;
	terrain.GetComponent(TerrainCollider).enabled = true;
}
